#| -*-Scheme-*-

$Id: mset.scm,v 1.2 1999/01/02 06:11:34 cph Exp $

Copyright (c) 1988, 1999 Massachusetts Institute of Technology

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
|#

;;;; Multiset Operations

(declare (usual-integrations))

(define (multiset->set multiset)
  (if (null? multiset)
      '()
      (let ((set (multiset->set (cdr multiset))))
	(if (memq (car multiset) set)
	    set
	    (cons (car multiset) set)))))

(define (multiset-difference x y)
  (if (null? y)
      x
      (let loop ((x x))
	(cond ((null? x) '())
	      ((memq (car x) y) (loop (cdr x)))
	      (else (cons (car x) (loop (cdr x))))))))