/* -*-C-*-

$Id: version.h,v 11.184 2002/06/17 05:19:36 cph Exp $

Copyright (c) 1988-2002 Massachusetts Institute of Technology

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA.
*/

/* This file contains version information for the microcode. */

/* Scheme system release version */

#ifndef SCHEME_RELEASE
#define SCHEME_RELEASE		"7.7.1"
#endif

/* Microcode release version */

#ifndef SCHEME_VERSION
#define SCHEME_VERSION		14
#endif
#ifndef SCHEME_SUBVERSION
#define SCHEME_SUBVERSION	9
#endif
