#| -*-Scheme-*-

$Id: files-core.scm,v 1.2 2000/12/05 21:23:50 cph Exp $

Copyright (c) 2000 Massachusetts Institute of Technology

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
|#

;;;; Core C files used on all platforms.

"artutl"
"avltree"
"bignum"
"bigprm"
"bitstr"
"boot"
"char"
"comutl"
"daemon"
"debug"
"dfloat"
"error"
"extern"
"fasload"
"fixnum"
"flonum"
"generic"
"hooks"
"hunk"
"intern"
"interp"
"intprm"
"list"
"lookprm"
"lookup"
"missing"
"obstack"
"option"
"osscheme"
"ostty"
"outf"
"prim"
"primutl"
"prmcon"
"ptrvec"
"purutl"
"regex"
"rgxprim"
"step"
"storage"
"string"
"syntax"
"sysprim"
"term"
"tterm"
"transact"
"utils"
"vector"
"wind"
