#| -*-Scheme-*-

$Id: files-optional.scm,v 1.3 2001/02/28 20:05:39 cph Exp $

Copyright (c) 2000-2001 Massachusetts Institute of Technology

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
|#

;;;; Optional C files that are conditionally linked in.

"cmpint"
"prbfish"
"prgdbm"
"prmcrypt"
"prmd5"
"prmhash"
"pruxdld"
"termcap"
"terminfo"
"tparam"
"x11base"
"x11color"
"x11graph"
"x11term"
