@echo off
rem MIT Scheme microcode configuration script for Windows NT
rem
rem Copyright (c) 1993 Massachusetts Institute of Technology
rem
rem $Id: setenv.bat,v 1.1 1993/07/18 20:31:17 gjr Exp $
rem
set path=%path%;c:\usr\lib\masm\bin
set include=%lib%;c:\usr\lib\masm\include
d:\mstools\setenv d:\mstools
